__name__ = "aiogoogle"
__about__ = "Async Google API client"
__description__ = __about__
__url__ = "https://github.com/omarryhan/aiogoogle"
__version_info__ = ("5", "2", "0")
__version__ = ".".join(__version_info__)
__author__ = "Omar Ryhan"
__author_email__ = "omarryhan@gmail.com"
__maintainer__ = "Omar Ryhan"
__license__ = "MIT"
__copyright__ = "(c) 2018 by Omar Ryhan"



__all__ = ["Aiogoogle"]

from contextvars import ContextVar
from typing import Optional

from .resource import GoogleAPI
from .auth.managers import Oauth2Manager, ApiKeyManager, OpenIdConnectManager, ServiceAccountManager
from .sessions.aiohttp_session import AiohttpSession
from .data import DISCOVERY_SERVICE_V1_DISCOVERY_DOC


# Discovery doc reference https://developers.google.com/discovery/v1/reference/apis

# Format string for a URL used to fetch a Google API discovery document from v2 of the Google Discovery Service
GOOGLE_DISCOVERY_SERVICE_URL_FMT_V2 = (
    "https://{api}.googleapis.com/$discovery/rest?version={api_version}"
)


class Aiogoogle:
    """
    Main entry point for Aiogoogle.

    This class acts as tiny wrapper around:

        1. Discovery Service v1 API
        2. Aiogoogle's OAuth2 manager
        3. Aiogoogle's API key manager
        4. Aiogoogle's OpenID Connect manager
        5. Aiogoogle's service account manager 
        6. One of Aiogoogle's implementations of a session object

    Arguments:

        session_factory (aiogoogle.sessions.abc.AbstractSession): AbstractSession Implementation. Defaults to ``aiogoogle.sessions.aiohttp_session.AiohttpSession``

        api_key (aiogoogle.auth.creds.ApiKey): Google API key
        
        user_creds (aiogoogle.auth.creds.UserCreds): OAuth2 cser credentials 

        client_creds (aiogoogle.auth.creds.ClientCreds): OAuth2 client credentials

        service_account_creds (aiogoogle.auth.creds.ServiceAccountCreds): Service account credentials
        
    Note: 
    
        In case you want to instantiate a custom session with initial parameters, you can pass an anonymous factory. e.g. ::
        
            >>> sess = lambda: Session(your_custome_arg, your_custom_kwarg=True)
            >>> aiogoogle = Aiogoogle(session_factory=sess)
    """

    def __init__(
        self,
        session_factory=AiohttpSession,
        api_key=None,
        user_creds=None,
        client_creds=None,
        service_account_creds=None,
    ):

        self.session_factory = session_factory
        # Guarantees that each context manager gets its own active_session.
        self.session_context: ContextVar[session_factory] = ContextVar("active_session", default=None)

        # Keys
        self.api_key = api_key
        self.user_creds = user_creds
        self.client_creds = client_creds
        self.service_account_creds = service_account_creds

        # Auth managers
        self.api_key_manager = ApiKeyManager(api_key=self.api_key)
        self.oauth2 = Oauth2Manager(
            self.session_factory, client_creds=self.client_creds
        )
        self.openid_connect = OpenIdConnectManager(
            self.session_factory, client_creds=self.client_creds
        )
        self.service_account_manager = ServiceAccountManager(
            self.session_factory, creds=self.service_account_creds
        )

        # Discovery service
        self.discovery_service = GoogleAPI(DISCOVERY_SERVICE_V1_DISCOVERY_DOC)

    # -------- Only 2 methods of Discover Service V1 ---------#

    async def list_api(self, name, preferred=None, fields=None):
        """
        https://developers.google.com/discovery/v1/reference/apis/list

        The discovery.apis.list method returns the list all APIs supported by the Google APIs Discovery Service.
        
        The data for each entry is a subset of the Discovery Document for that API, and the list provides a directory of supported APIs.
        
        If a specific API has multiple versions, each of the versions has its own entry in the list.

        Example:

            ::

                >>> await aiogoogle.list_api('youtube')

                {
                    "kind": "discovery#directoryList",
                    "discoveryVersion": "v1",
                    "items": [
                        {
                            "kind": "discovery#directoryItem",
                            "id": "youtube:v3",
                            "name": "youtube",
                            "version": "v3",
                            "title": "YouTube Data API",
                            "description": "Supports core YouTube features, such as uploading videos, creating and managing playlists, searching for content, and much more.",
                            "discoveryRestUrl": "https://www.googleapis.com/discovery/v1/apis/youtube/v3/rest",
                            "discoveryLink": "./apis/youtube/v3/rest",
                            "icons": {
                                "x16": "https://www.google.com/images/icons/product/youtube-16.png",
                                "x32": "https://www.google.com/images/icons/product/youtube-32.png"
                            },
                            "documentationLink": "https://developers.google.com/youtube/v3",
                            "preferred": true
                        }
                    ]
                }

        Arguments:

            name (str): Only include APIs with the given name.

            preferred (bool): Return only the preferred version of an API.  "false" by default.

            fields (str): Selector specifying which fields to include in a partial response.

        Returns:

            dict:

        Raises:

            aiogoogle.excs.HTTPError
        """

        request = self.discovery_service.apis.list(
            name=name, preferred=preferred, fields=fields
        )
        return await self.as_anon(request)

    async def discover(self, api_name, api_version=None, validate=False, *, disco_doc_ver: Optional[int] = None):
        """
        Donwloads a discovery document from Google's Discovery Service V1 and sets it a ``aiogoogle.resource.GoogleAPI``

        Note:

            It is recommended that you explicitly specify an API version.

            When you leave the API version as None, Aiogoogle uses the ``list_api`` method to search for the best fit version of the given API name.

            This will result in sending two http requests instead of just one.

        Arguments:

            api_name (str): API name to discover. *e.g.: "youtube"*

            api_version (str): API version to discover *e.g.: "v3" not "3" and not 3*

            validate (bool): Set this to True to use this lib's built in parameter validation logic. Note that you shouldn't rely on this for critical user input validation.
            disco_doc_ver: Specify which Google Discovery Service version to fetch a discovery document with.
                Useful for fetching discovery docs for Google APIs that aren't supported by
                the default version of the Google Discovery Service Aiogoogle uses.

        Returns:

            aiogoogle.resource.GoogleAPI: An object that will then be used to create API requests

        Raises:

            aiogoogle.excs.HTTPError

        """

        if api_version is None:
            # Search for name in self.list_api and return best match
            discovery_list = await self.list_api(api_name, preferred=True)

            if discovery_list["items"]:
                api_name = discovery_list["items"][0]["name"]
                api_version = discovery_list["items"][0]["version"]
            else:
                raise ValueError("Invalid API name")

        request = self.discovery_service.apis.getRest(
            api=api_name, version=api_version, validate=False
        )

        # Allow clients to fetch Google API discovery docs from v2 of the Google Discovery Service
        if disco_doc_ver == 2:
            request.url = GOOGLE_DISCOVERY_SERVICE_URL_FMT_V2.format(
                api=api_name, api_version=api_version
            )

        discovery_document = await self.as_anon(request)

        return GoogleAPI(discovery_document, validate)

    # -------- Send Requests ----------#

    async def as_user(self, *requests, timeout=None, full_res=False, user_creds=None, raise_for_status=True):
        """ 
        Sends requests on behalf of ``self.user_creds`` (OAuth2)
        
        Arguments:

            *requests (aiogoogle.models.Request):

                Requests objects typically created by ``aiogoogle.resource.Method.__call__``

            timeout (int):

                Total timeout for all the requests being sent

            full_res (bool):

                If True, returns full HTTP response object instead of returning it's content

            user_creds (aiogoogle.auth.creds.UserCreds):

                If you pass user_creds here, they will only be used for this one request.

            raise_for_status (bool):

                If True, raises an HTTP error on HTTP status codes >= 400

        Returns:

            aiogoogle.models.Response:
        """
        user_creds = user_creds or self.user_creds
        if user_creds is None:
            raise TypeError("No user credentials were found")

        # Refresh credentials
        if (
            user_creds.get("expires_at") is None
            or (user_creds.get("expires_at") and self.oauth2.is_expired(user_creds) is True)
        ):
            user_creds = await self.oauth2.refresh(
                user_creds, client_creds=self.client_creds
            )

            # Set refreshed user_creds if ones were already existing
            if self.user_creds is not None:
                self.user_creds = user_creds

        authorized_requests = [
            self.oauth2.authorize(request, user_creds) for request in requests
        ]

        return await self.send(
            *authorized_requests,
            timeout=timeout,
            full_res=full_res,
            raise_for_status=raise_for_status,
            session_factory=self.session_factory,
            auth_manager=self.oauth2
        )

    async def as_service_account(
            self, *requests, timeout=None, full_res=False, service_account_creds=None, raise_for_status=True):
        """ 
        Sends requests on behalf of ``self.user_creds`` (OAuth2)
        
        Arguments:

            *requests (aiogoogle.models.Request):

                Requests objects typically created by ``aiogoogle.resource.Method.__call__``

            timeout (int):

                Total timeout for all the requests being sent

            full_res (bool):

                If True, returns full HTTP response object instead of returning it's content

            service_account_creds (aiogoogle.auth.creds.ServiceAccountCreds):

                You only have to pass ``service_account_creds`` once, either here or when instantiating an instance of Aiogoogle.

            raise_for_status (bool):

                If True, raises an HTTP error on HTTP status codes >= 400

        Returns:

            aiogoogle.models.Response:
        """
        service_account_creds = service_account_creds or self.service_account_creds
        if service_account_creds is None:
            raise TypeError("Please pass service account creds")

        await self.service_account_manager.refresh()

        authorized_requests = [
            self.service_account_manager.authorize(request) for request in requests
        ]

        return await self.send(
            *authorized_requests,
            timeout=timeout,
            full_res=full_res,
            raise_for_status=raise_for_status,
            session_factory=self.session_factory,
            auth_manager=self.service_account_manager,
        )

    async def as_api_key(self, *requests, timeout=None, full_res=False, api_key=None, raise_for_status=True):
        """ 
        Sends requests on behalf of ``self.api_key`` (OAuth2)
        
        Arguments:

            *requests (aiogoogle.models.Request):

                Requests objects typically created by ``aiogoogle.resource.Method.__call__``

            timeout (int):

                Total timeout for all the requests being sent

            full_res (bool):

                If True, returns full HTTP response object instead of returning it's content

            api_key (string):

                If you pass an API key here, it will only be used for this one request.

            raise_for_status (bool):

                If True, raises an HTTP error on HTTP status codes >= 400

        Returns:

            aiogoogle.models.Response:
        """
        api_key = api_key or self.api_key
        if api_key is None:
            raise TypeError("Please pass an API key")

        authorized_requests = [
            self.api_key_manager.authorize(request, api_key)
            for request in requests
        ]

        return await self.send(
            *authorized_requests,
            timeout=timeout,
            full_res=full_res,
            raise_for_status=raise_for_status,
            session_factory=self.session_factory,
            auth_manager=self.api_key_manager
        )

    async def as_anon(self, *requests, timeout=None, full_res=False, raise_for_status=True):
        """ 
        Sends unauthorized requests
        
        Arguments:

            *requests (aiogoogle.models.Request):

                Requests objects typically created by ``aiogoogle.resource.Method.__call__``

            timeout (int):

                Total timeout for all the requests being sent

            full_res (bool):

                If True, returns full HTTP response object instead of returning it's content

            raise_for_status (bool):

                If True, raises an HTTP error on HTTP status codes >= 400

        Returns:

            aiogoogle.models.Response:
        """
        return await self.send(
            *requests,
            timeout=timeout,
            full_res=full_res,
            raise_for_status=raise_for_status,
            session_factory=self.session_factory,
            auth_manager=None
        )

    def _get_session(self):
        return self.session_context.get()

    def _set_session(self):
        session = self.session_factory()
        self.session_context.set(session)
        return session

    async def send(self, *args, **kwargs):
        session = self._get_session()
        if session is None:
            session = self._set_session()
        return await session.send(*args, **kwargs)

    async def __aenter__(self):
        session = self._get_session()
        if session is None:
            session = self._set_session()
            await session.__aenter__()
            return self
        raise RuntimeError("Nesting context managers using the same Aiogoogle object is not allowed.")

    async def __aexit__(self, *args):
        session = self._get_session()
        await session.__aexit__(*args)
        # Had to add this because there's no use of keeping a closed session
        # Closed sessions cannot be reopened, so it's better to just get rid of the object
        self.session_context.set(None)



__all__ = ["ApiKey", "UserCreds", "ClientCreds", "ServiceAccountCreds"]


from ..utils import _dict


class ApiKey(str):
    pass


class IdToken(_dict):
    """
    OpenID connect ID token

    Example:

        {"iss":"accounts.google.com",
        "at_hash":"HK6E_P6Dh8Y93mRNtsDB1Q",
        "email_verified":"true",
        "sub":"10769150350006150715113082367",
        "azp":"1234987819200.apps.googleusercontent.com",
        "email":"jsmith@example.com",
        "aud":"1234987819200.apps.googleusercontent.com",
        "iat":1353601026,
        "exp":1353604926,
        "nonce": "0394852-3190485-2490358",
        "hd":"example.com"}

    Attributes:

        iss (str):

            * PROVIDED: always
        
            * The Issuer Identifier for the Issuer of the response. Always https://accounts.google.com or accounts.google.com for Google ID tokens.
        
        at_hash	(str):
        
            * Access token hash.
            
            * Provides validation that the access token is tied to the identity token.
            
            * If the ID token is issued with an access token in the server flow, this is always included.
            
            * This can be used as an alternate mechanism to protect against cross-site request forgery attacks, but if you follow Step 1 and Step 3 it is not necessary to verify the access token.
        
        email_verified (bool)
        
            * True if the user's e-mail address has been verified; otherwise false.
        
        sub (str):

            * PROVIDED: always
        
            * An identifier for the user, unique among all Google accounts and never reused.
            
            * A Google account can have multiple emails at different points in time, but the sub value is never changed.
            
            * Use sub within your application as the unique-identifier key for the user.
        
        azp	(str):
        
            * The client_id of the authorized presenter.
            
            * This claim is only needed when the party requesting the ID token is not the same as the audience of the ID token.
            
            * This may be the case at Google for hybrid apps where a web application and Android app have a different client_id but share the same project.
        
        email (str):
        
            * The user's email address.
            
            * This may not be unique and is not suitable for use as a primary key.
            
            * Provided only if your scope included the string "email".
        
        profile (str):
        
            * The URL of the user's profile page. Might be provided when:
        
                * The request scope included the string "profile"
        
                * The ID token is returned from a token refresh
        
            * When profile claims are present, you can use them to update your app's user records. 
            
            * Note that this claim is never guaranteed to be present.
        
        picture (str):
        
            * The URL of the user's profile picture. Might be provided when:
        
                * The request scope included the string "profile"
        
                * The ID token is returned from a token refresh
        
            * When picture claims are present, you can use them to update your app's user records.
            
            * Note that this claim is never guaranteed to be present.
        
        name (str):
        
            * The user's full name, in a displayable form. Might be provided when:
        
                * The request scope included the string "profile"
        
                * The ID token is returned from a token refresh
        
            * When name claims are present, you can use them to update your app's user records. Note that this claim is never guaranteed to be present.
        
        aud (str):
        
            * PROVIDED:	always
            
            * The audience that this ID token is intended for.
            
            * It must be one of the OAuth 2.0 client IDs of your application.
        
        iat (str):
        
            * PROVIDED: always
            
            * The time the ID token was issued, represented in Unix time (integer seconds).
        
        exp (str):
        
            * PROVIDED: always
            
            * The time the ID token expires, represented in Unix time (integer seconds).
        
        nonce (str): 
        
            * The value of the nonce supplied by your app in the authentication request.
            
            * You should enforce protection against replay attacks by ensuring it is presented only once.
        
        hd (str):
        
            * The hosted G Suite domain of the user. Provided only if the user belongs to a hosted domain.
    """

    pass


class UserCreds(_dict):
    """
    OAuth2 User Credentials Dictionary

    Attributes:

        access_token (str): Access Token
        refresh_token (str): Refresh Token
        expires_in (int): seconds till expiry from creation
        expires_at (str): JSON datetime ISO 8601 expiry datetime
        scopes (list): list of scopes owned by access token
        
        id_token (aiogoogle.auth.creds.IdToken): Decoded OpenID JWT
        id_token_jwt (str): Encoded OpenID JWT
        
        token_type (str): Bearer
        token_uri (str): URI where this token was issued from
        token_info_uri (str): URI where one could get more info about this token
        revoke_uri (str): URI where this token should be revoked
        
    """

    def __init__(
        self,
        access_token=None,
        refresh_token=None,
        expires_in=None,
        expires_at=None,
        scopes=None,
        id_token=None,
        id_token_jwt=None,
        token_type=None,
        token_uri=None,
        token_info_uri=None,
        revoke_uri=None,
    ):
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.expires_in = expires_in
        self.expires_at = expires_at
        self.scopes = scopes

        self.id_token = id_token
        self.id_token_jwt = id_token_jwt

        self.token_type = token_type
        self.token_uri = token_uri
        self.token_info_uri = token_info_uri
        self.revoke_uri = revoke_uri


class ClientCreds(_dict):
    """
    OAuth2 Client Credentials Dictionary

    Examples:

        Scopes: ['openid', 'email', 'https://www.googleapis.com/auth/youtube.force-ssl']

    Attributes:

        client_id (str): OAuth2 client ID
        client_secret (str): OAuth2 client secret
        scopes (list): List of scopes that this client should request from resource server
        redirect_uri (str): client's redirect uri
    """

    def __init__(
        self, client_id=None, client_secret=None, scopes=None, redirect_uri=None
    ):
        self.client_id = client_id
        self.client_secret = client_secret
        self.scopes = scopes
        self.redirect_uri = redirect_uri


class ServiceAccountCreds(_dict):
    """
    Service account key (the one you download from Google Cloud Console) +
    some additional optional attributes.
    If you have created the service account using Google's IAM REST API, rather
    than downloading it using gcloud or Google's cloud console, the downloaded
    JSON key file will look different. Check here:
    https://cloud.google.com/iam/docs/creating-managing-service-account-keys#creating_service_account_keys
    to be able to pass the correct info to this data model.

    Attributes:

        type (string): Should be "service_account"

        project_id (string): Your project ID

        private_key_id (string): Private key ID

        private_key (string): RSA Private key

        client_email (string): service account email

        client_id (string): service account ID

        auth_uri (string): Auth URI

        token_uri (string): Token URI

        auth_provider_x509_cert_url (string): Auth provider cert URL

        client_x509_cert_url (string): Cert URL

        scopes (Sequence[str]): Scopes to request during the authorization grant.

        subject (str): For domain-wide delegation, the email address of the user to for which to request delegated access.

        additional_claims (Mapping[str, str]): Any additional claims for the JWT assertion used in the authorization grant.

    Example:

        ::

            service_account_creds = ServiceAccountCreds(
                scopes=["https://www.googleapis.com/auth/cloud-platform"],
                **json.load(open('service-account-key.json'))
            )

            # or

            service_account_creds = {
                "scopes": ["..."],
                **json.load(open('service-account-key.json'))
            }
    """
    def __init__(
        self, type=None, project_id=None, private_key_id=None, private_key=None,
        client_email=None, client_id=None, auth_uri=None, token_uri=None, auth_provider_x509_cert_url=None,
        client_x509_cert_url=None, subject=None, scopes=None, additional_claims=None
    ):
        self.type = type
        self.project_id = project_id
        self.private_key_id = private_key_id
        self.private_key = private_key
        self.client_email = client_email
        self.client_id = client_id
        self.auth_uri = auth_uri
        self.token_uri = token_uri
        self.auth_provider_x509_cert_url = auth_provider_x509_cert_url
        self.client_x509_cert_url = client_x509_cert_url
        # Starting from here, the info is not included in the key file
        # and is optional. (Some of the info above is also optional).
        self.subject = subject
        self.scopes = scopes
        self.additional_claims = additional_claims or {}



__all__ = ["AiohttpSession"]

import asyncio
from json import JSONDecodeError

from aiohttp import ClientSession, MultipartWriter
from aiohttp.client_exceptions import ContentTypeError
import aiofiles
from aiofiles import os as async_os
import async_timeout

from ..models import Response
from .abc import AbstractSession


async def _get_file_size(full_file_path):
    stat = await async_os.stat(full_file_path)
    return stat.st_size


async def _aiter_file(file_name, chunk_size):
    """ Async file generator """
    async with aiofiles.open(file_name, "rb") as f:
        chunk = await f.read(chunk_size)
        while chunk:
            yield chunk
            chunk = await f.read(chunk_size)


async def _read_file(file_name):
    async with aiofiles.open(file_name, "rb") as file:
        return await file.read()


class AiohttpSession(ClientSession, AbstractSession):
    async def send(
        self,
        *requests,
        timeout=None,
        full_res=False,
        raise_for_status=True,
        session_factory=None,
        auth_manager=None
    ):
        async def resolve_response(request, response):
            data = None
            json = None
            download_file = None
            upload_file = None
            pipe_from = None
            pipe_to = None

            # If downloading file:
            if request.media_download:
                chunk_size = request.media_download.chunk_size
                download_file = request.media_download.file_path
                pipe_to = request.media_download.pipe_to

                if pipe_to:
                    async for line in response.content.iter_chunked(chunk_size):
                        await pipe_to.write(line)

                if download_file:
                    async with aiofiles.open(download_file, "wb+") as f:
                        async for line in response.content.iter_chunked(chunk_size):
                            await f.write(line)
            else:
                if response.status != 204:  # If no (no content)
                    try:
                        json = await response.json()
                    except (JSONDecodeError, ContentTypeError):
                        try:
                            data = await response.text()
                        except ContentTypeError:
                            try:
                                data = await response.read()
                            except ContentTypeError:
                                data = None

            if request.media_upload:
                upload_file = request.media_upload.file_path

            return Response(
                url=str(response.url),
                headers=response.headers,
                status_code=response.status,
                json=json,
                data=data,
                reason=response.reason if getattr(response, "reason") else None,
                req=request,
                download_file=download_file,
                pipe_to=pipe_to,
                upload_file=upload_file,
                pipe_from=pipe_from,
                session_factory=session_factory,
                auth_manager=auth_manager
            )

        async def fire_request(request):
            request.headers["Accept-Encoding"] = "gzip"
            request.headers["User-Agent"] = "Aiogoogle Aiohttp (gzip)"
            if request.media_upload:
                # Validate
                await request.media_upload.run_validation(_get_file_size)

                # If multipart pass a file async generator
                if request.media_upload.multipart is True:
                    with MultipartWriter('mixed') as mpwriter:
                        mpwriter.append_json(request.json)

                        mpwriter.append(
                            request.media_upload.aiter_file(_aiter_file),
                            headers={"Content-Type": request.upload_file_content_type} if request.upload_file_content_type else None
                        )

                        req_content_type = 'multipart/related'

                        request.headers.update({"Content-Type": f"{req_content_type}; boundary={mpwriter.boundary}"})

                        return await self.request(
                            method=request.method,
                            url=request.media_upload.upload_path,
                            headers=request.headers,
                            data=mpwriter,
                            timeout=request.timeout,
                            verify_ssl=request._verify_ssl,
                        )
                # Else load file to memory and send
                else:
                    read_file = await request.media_upload.read_file(_read_file)
                    if request.upload_file_content_type:
                        request.headers.update({"Content-Type": request.upload_file_content_type})
                    return await self.request(
                        method=request.method,
                        url=request.media_upload.upload_path,
                        headers=request.headers,
                        data=read_file,
                        json=request.json,
                        timeout=request.timeout,
                        verify_ssl=request._verify_ssl,
                    )
            # Else, if no file upload
            else:
                return await self.request(
                    method=request.method,
                    url=request.url,
                    headers=request.headers,
                    data=request.data,
                    json=request.json,
                    timeout=request.timeout,
                    verify_ssl=request._verify_ssl,
                )

        # ----------------- send sequence ------------------#
        async def get_response(request):
            response = await fire_request(request)
            response = await resolve_response(request, response)
            if raise_for_status is True:
                response.raise_for_status()
            return response

        async def get_content(request):
            response = await get_response(request)
            return response.content

        # ----------------- /send sequence ------------------#

        async def schedule_tasks():
            if full_res is True:
                tasks = [
                    asyncio.ensure_future(get_response(request)) for request in requests
                ]
            else:
                tasks = [
                    asyncio.ensure_future(get_content(request)) for request in requests
                ]
            return await asyncio.gather(*tasks, return_exceptions=False)

        session_factory = self.__class__ if session_factory is None else session_factory

        if timeout is not None:
            async with async_timeout.timeout(timeout):
                results = await schedule_tasks()
        else:
            results = await schedule_tasks()

        return (
            results[0] if isinstance(results, list) and len(results) == 1 else results
        )



__all__ = ["create_secret"]

import hashlib
import os
import datetime
import sys

from ..utils import _parse_isoformat


def create_secret(bytes_length=1024):  # pragma: no cover
    # return hashlib.sha256(os.urandom(bytes_length << 3)).hexdigest()
    return hashlib.sha256(os.urandom(bytes_length)).hexdigest()


def _get_expires_at(expires_in):
    expires_at = datetime.datetime.utcnow() + datetime.timedelta(seconds=expires_in)
    # account for clock skew
    expires_at -= datetime.timedelta(seconds=120)
    return expires_at.isoformat()


def _is_expired(expires_at):
    # Refresh in case there's no expires_at present
    if expires_at is None:
        return True
    if not isinstance(expires_at, datetime.datetime):
        # datetime.fromisoformat is 3.7+
        if sys.version_info[1] <= 6:
            expires_at = _parse_isoformat(expires_at)
        else:
            expires_at = datetime.datetime.fromisoformat(expires_at)
    if datetime.datetime.utcnow() >= expires_at:
        return True
    else:
        return False



__all__ = ["GoogleAPI", "Resource", "Method"]

import re
import warnings
from urllib.parse import urlencode, quote
from functools import wraps
from typing import List, Generic, TypeVar

from .excs import ValidationError
from .utils import _safe_getitem
from .models import MediaDownload, MediaUpload, ResumableUpload, Request
from .validate import validate as validate_


T = TypeVar("T")  # Generic type var

# These are the hard-coded kwargs in Method.__call__
# They're used for testing whether those names will collide with any of the url parameters that are provided by any of the discovery docs.
# If collisions were to be found, that would mean that the user won't be able to pass a url_parameter that shares the same name with any of the RESERVED_KEYWORDS.
RESERVED_KEYWORDS = [
    "validate",
    "data",
    "json",
    "upload_file",
    "pipe_from",
    "download_file",
    "pipe_to",
    "timeout",
]

# From: https://github.com/googleapis/google-api-python-client/blob/master/googleapiclient/discovery.py
# Parameters accepted by the stack, but not visible via discovery.
STACK_QUERY_PARAMETERS = frozenset(["trace", "pp", "strict"])
STACK_QUERY_PARAMETER_DEFAULT_VALUE = {"type": "string", "location": "query"}
MEDIA_SIZE_BIT_SHIFTS = {"KB": 10, "MB": 20, "GB": 30, "TB": 40}

# TODO: etagRequired: {
#    type: "boolean",  # noqa: F821 (weird error)
#    description: "Whether this method requires an ETag to be specified. The ETag is sent as an HTTP If-Match or If-None-Match header."
#    }
# NOTE: etagRequired is only mentioned once in all of the discovery documents available from Google. (In discovery_service-v1. So, it isn't actually being used)


def _temporarily_add_back_dashes_to_param_definitions(f):
    """
        When instantiating a Method, Method's constructor will remove all 
        dashes from the names of its URI params and global params in order
        to make it possible to pass uri params through function calls
        e.g. this is viable get_videos(my_videos=True)
        this is will fail get_videos(my-videos=True)

        This function momentarily adds back '-' to url parameters and passed uri_params
        in order to be processed and validated correctly and comply with the disc doc
        Reverts back to '_' after wrapped function is done
    """

    @wraps(f)
    def wrapper(
        self,
        validate=None,
        data=None,
        json=None,
        upload_file=None,
        pipe_from=None,
        download_file=None,
        pipe_to=None,
        timeout=None,
        **uri_params,
    ):
        # unfix urls
        uri_params = self._replace_dashes_with_underscores_in_user_provided_params(
            uri_params, self.parameters
        )

        # unfix params
        self._method_specs["parameters"] = self._restore_dashes_to_params(
            self._method_specs.get("parameters")
        )
        self._global_parameters = self._restore_dashes_to_params(
            self._global_parameters
        )

        # Run function
        results = f(
            self,
            validate,
            data,
            json,
            upload_file,
            pipe_from,
            download_file,
            pipe_to,
            timeout,
            **uri_params,
        )

        # fix params again
        self._method_specs[
            "parameters"
        ] = self._replace_dashes_with_underscores_in_params(
            self._method_specs.get("parameters")
        )
        self._global_parameters = self._replace_dashes_with_underscores_in_params(
            self._global_parameters
        )

        return results

    return wrapper


class Method:
    def __init__(
        self,
        name,
        method_specs,
        global_parameters,
        schemas,
        root_url,
        service_path,
        batch_path,
        validate,
    ):
        # Replaces '-'s with '_'s and preserve old names to revert back to them after this method is called
        global_parameters = self._replace_dashes_with_underscores_in_params(
            global_parameters
        )
        method_specs["parameters"] = self._replace_dashes_with_underscores_in_params(
            method_specs.get("parameters")
        )

        self.name = name
        self._method_specs = method_specs
        self._global_parameters = global_parameters
        self._schemas = schemas

        self._root_url = root_url
        self._service_path = service_path
        self._batch_path = batch_path

        if (
            self["useMediaDownloadService"] is True
            and self["supportsMediaDownload"] is True
        ):
            self._download_base_url = self._root_url + "download/" + self._service_path
        else:
            self._download_base_url = None

        self._base_url = self._root_url + self._service_path
        self._batch_url = self._root_url + self._batch_path

        self._should_validate = validate

    # ---- Changes URL parameters with a "-" to "_" -----#
    # Depends on how you view it, but this section also changes robots with small mouths to robots with big mouths

    @staticmethod
    def _replace_dashes_with_underscores_in_params(param_set: Generic[T]) -> T:
        if param_set:
            for name, schema in list(param_set.items()):
                if "-" in name:
                    new_name = name.replace("-", "_")  # See?!
                    schema["orig_name"] = name
                    param_set[new_name] = schema
                    del param_set[name]
        return param_set

    @staticmethod
    def _restore_dashes_to_params(param_set: Generic[T]) -> T:
        if param_set:
            for name, schema in list(param_set.items()):
                if "orig_name" in schema:
                    param_set[schema["orig_name"]] = schema
                    del param_set[name]
        return param_set

    @staticmethod
    def _replace_dashes_with_underscores_in_user_provided_params(
        uri_params: Generic[T], parameters
    ) -> T:
        for k, v in list(uri_params.items()):
            if "_" in k:
                if k in parameters:
                    if "orig_name" in parameters[k]:
                        uri_params[parameters[k]["orig_name"]] = v
                        del uri_params[k]
        return uri_params

    @staticmethod
    def _del_none_params(uri_params: Generic[T]) -> T:
        for k, v in list(uri_params.items()):
            if v is None:
                del uri_params[k]
        return uri_params

    # ---- / Changes URL parameters with a "-" to "_" -----#

    @property
    def request(self) -> dict:
        """ Returns expected request body """
        body = self["request"]
        if body.get("$ref"):
            return self._schemas.get(body["$ref"])
        else:
            return body

    @property
    def response(self) -> dict:
        """ Retruns expected response body """
        body = self["response"]
        if body.get("$ref"):
            return self._schemas.get(body["$ref"])
        else:
            return body

    @property
    def parameters(self) -> dict:
        """ 
        Parameters property

        Returns:

            dict: All parameters that this method can take as described in the discovery document
        """
        if not self._global_parameters and not self["parameters"]:
            return {}
        elif not self._global_parameters:
            return self["parameters"]
        elif not self["parameters"]:
            return self._global_parameters
        else:
            return {**self["parameters"], **self._global_parameters}

    @property
    def optional_parameters(self) -> List[str]:
        """
        Optional Parameters

        Returns:

            list: List of the names of optional parameters this method takes
        """
        return (
            [
                parameter_name
                for parameter_name, parameter_info in self.parameters.items()
                if parameter_info.get("required") is not True
            ]
            if self.parameters
            else []
        )

    @property
    def required_parameters(self) -> List[str]:
        """
        Required Parameters

        Returns:

            list: List of the names of required parameters this method takes
        """
        return (
            [
                parameter_name
                for parameter_name, parameter_info in self.parameters.items()
                if parameter_info.get("required") is True
            ]
            if self.parameters
            else []
        )

    @property
    def path_parameters(self) -> List[str]:
        """
        Path Parameters

        Returns:

            list: List of the names of path parameters this method takes
        """
        return (
            [
                param_name
                for param_name, param_info in self.parameters.items()
                if param_info.get("location") == "path"
            ]
            if self.parameters
            else []
        )

    @property
    def query_parameters(self) -> List[str]:
        """
        Query Parameters

        Returns:

            list: List of the names of Query parameters this method takes
        """
        return (
            [
                param_name
                for param_name, param_info in self.parameters.items()
                if param_info.get("location") == "query"
            ]
            if self.parameters
            else []
        )

    @property
    def required_query_parameters(self) -> List[str]:
        """
        Required Query Parameters

        Returns:

            list: List of the names of required query parameters this method takes
        """
        return (
            [
                param_name
                for param_name in self.query_parameters
                if param_name in self.required_parameters
            ]
            if self.query_parameters
            else []
        )

    @property
    def optional_query_parameters(self) -> List[str]:
        """
        Optional Query Parameters

        Returns:

            list: List of the names of optional query parameters this method takes
        """
        return (
            [
                param_name
                for param_name in self.query_parameters
                if param_name in self.optional_parameters
            ]
            if self.query_parameters
            else []
        )

    def __getitem__(self, key):
        """
        Examples:

            ::

                >>> self['description'] 
                
                "method description"

                >>> self['scopes']
                
                ['returns', 'scopes', 'required', 'by', 'this', 'method', 'in', 'a', 'list']

                >>> self['supportsMediaDownload']
                
                False

                >>> self['supportsMediaUpload']

                True

                >>> self['httpMethod']

                'GET'

        Hint:


            Using this method with ``scopes`` as an argument can be useful for incremental authorization. (Requesting scopes when needed. As opposed to requesting them at once)

            for more: https://developers.google.com/identity/protocols/OAuth2WebServer#incrementalAuth


        """
        return self._method_specs.get(key)

    def _validate(self, instance, schema, schema_name=None):
        return validate_(instance, schema, self._schemas, schema_name)

    @_temporarily_add_back_dashes_to_param_definitions
    def __call__(
        self,
        validate=None,
        data=None,
        json=None,
        upload_file=None,
        pipe_from=None,
        download_file=None,
        pipe_to=None,
        timeout=None,
        path_params_safe_chars={},
        **uri_params,
    ) -> Request:
        """ 
        Builds a request from this method

        Note:

            * When passing ``datetime.datetime or datetime.date`` pass them in json format.
            
            * Aiogoogle won't do that as it would be a big hassle to iterate over every item in ``*uri_params``, ``json`` and ``data`` to check if there's any datetime objects.

            * Fortunately Python makes it really easy to achieve that. 
            
            * Instead of passing say ``datetime.datetime.utcnow()``, pass: ``datetime.datetime.utcnow().jsonformat()``

        Note:

            * All ``None`` values are ommited before sending to Google apis, if you want to explicitly pass a JSON null then pass it as ``"null"`` not ``None`` 

        Arguments:

            validate (bool): Overrides :param: aiogoogle.Aiogoole.validate if not None
            
            json (dict): Json body
            
            data (any): Data body (Bytes, text, www-url-form-encoded and others)
            
            upload_file (str): full path of file to upload
            
            download_file (str): full path of file to download to
            
            timeout (str): total timeout for this request
            
            path_params_safe_chars (dict): Dictionary of safe characters for each path parameter.
            
            **uri_params (dict): path and query, required and optional parameters

        Returns:

            aiogoogle.models.Request: An unsent request object
        """
        # If collisions are found between the 'key' of **uri_params and explicit kwargs e.g. data, json etc., then
        # priority will be given to explicit kwargs. With that said, it's not likely there will be any.
        # If you want to double check if there are any collisions,
        # you can append the API name and version you're using to tests.globals.ALL_APIS (only if they don't exist, otherwise, you shouldn't worry about collisions)
        # Then, run the unit tests and monitor: tests.test_discovery_document.test_parameters_not_colliding_with_google_api__call__ for failure

        #
        # NOTE: Use '_' instead of '-' when passing uri parameters that have a '-' in their names
        #

        # Remove params that are None
        uri_params = self._del_none_params(uri_params)

        # Assert timeout is int
        if timeout is not None:
            if not isinstance(timeout, int) or type(timeout) == bool:
                raise TypeError("Timeouts can only be ints or None")

        # Resolve validation status
        if not isinstance(validate, bool):
            validate = self._should_validate

        base_url = self._base_url

        # Build full url minus query & fragment
        url = self._build_url(
            base_url=base_url, uri_params=uri_params.copy(), validate=validate, path_params_safe_chars=path_params_safe_chars
        )

        # Filter out query parameters from all uri_params that were passed to this method
        passed_query_params = {
            param_name: param_info
            for param_name, param_info in uri_params.items()
            if param_name in self.query_parameters
        }

        # Ensure all required query parameteters were passed
        for param in self.required_query_parameters:
            if param not in passed_query_params:
                raise ValidationError(f'Missing query parameter: "{param}"')

        # Validate url query params
        if validate is True:
            if passed_query_params:
                for param_name, passed_param in passed_query_params.items():
                    schema = self.parameters[param_name]
                    if schema.get('repeated') and isinstance(passed_param, (list, set, tuple)):
                        for param in passed_param:
                            self._validate(
                                param,
                                schema,
                                schema_name=param_name
                            )
                    else:
                        self._validate(
                            passed_param,
                            schema,
                            schema_name=param_name,
                        )

        # Join query params
        if passed_query_params:
            uri = url + "?" + urlencode(passed_query_params, True)
        else:
            uri = url

        # Pop query uri_params consumed
        for param_name, _ in passed_query_params.items():
            del uri_params[param_name]

        # Pop uri_params consumed
        # for param_name in self["parameterOrder"]:
        for param_name in self.path_parameters:
            if param_name in uri_params:
                del uri_params[param_name]

        # Warn if not all uri_params were consumed/popped
        if uri_params:  # should be empty by now
            # If there's room for addtionalProperties, validate and add them to the URI
            if validate:
                if self.parameters.get("additionalProperties"):
                    for _, v in uri_params.items():
                        self._validate(
                            v,
                            self.parameters["additionalProperties"],
                            schema_name="Additional Url Parameters",
                        )
                else:
                    raise ValidationError(
                        f"Invalid (extra) parameters: {uri_params} were passed"
                    )
            else:
                if not self.parameters.get("additionalProperties"):
                    warnings.warn(
                        "Parameters {} were found and they're probably of no use."
                        " Check if they're valid parameters".format(str(uri_params))
                    )
            if "?" not in uri:
                uri = uri + "?" + urlencode(uri_params)
            else:
                uri = uri + "&" + urlencode(uri_params)

        # Ensure only one param for http req body.
        if json and data:
            raise TypeError(
                "Pass either json or data for the body of the request, not both."
                '\nThis is similar to the "body" argument in google-python-client'
                "\nThis will validate agains the $request key in this method's "
                "specs"
            )  # This raises a TypeError instead of a ValidationError because
            # it will probably make your session raise an error if it passes.
            # Better raise it early on

        # Validate body
        if validate is True:
            body = (
                json if json is not None else data if data is not None else None
            )  # json or data or None
            if body is not None:
                self._validate_body(body)

        # Validate selected options
        if download_file and pipe_to:
            raise ValueError(
                "can't have both (download_file) and (pipe_to) options"
            )

        if not download_file and not pipe_to:
            media_download = None

        # Process download_file
        if download_file:
            if validate is True:
                if self.__getitem__("supportsMediaDownload") is not True:
                    raise ValidationError(
                        "download_file was provided while method doesn't support media download"
                    )
            media_download = MediaDownload(file_path=download_file)

        # Process pipe_to
        if pipe_to:
            if validate is True:
                if self.__getitem__("supportsMediaDownload") is not True:
                    raise ValidationError(
                        "pipe_to was provided while method doesn't support media download"
                    )
            media_download = MediaDownload(pipe_to=pipe_to)

        if not upload_file and not pipe_from:
            media_upload = None

        # Process upload_file
        if upload_file:
            media_upload = self._build_upload_media(
                upload_file, uri, validate, fallback_url=url, pipe_from=None
            )

        # Process pipe_from
        if pipe_from:
            media_upload = self._build_upload_media(
                upload_file, uri, validate, fallback_url=url, pipe_from=pipe_from
            )

        return Request(
            method=self["httpMethod"],
            url=uri,
            batch_url=self._batch_url,
            data=data,
            json=json,
            timeout=timeout,
            media_download=media_download,
            media_upload=media_upload,
            callback=lambda res: res,  # TODO: get rid of this sorcery.
        )

    def _build_url(self, base_url, uri_params, validate, path_params_safe_chars):
        if self.path_parameters:
            # sort path params as sepcified in method_specs.parameterOrder
            sorted_required_path_params = (
                {}
            )  # Dict order is guaranteed (by insertion) as of Python 3.6
            for param_name in self["parameterOrder"]:
                try:
                    sorted_required_path_params[param_name] = uri_params.pop(param_name)
                except KeyError:
                    raise ValidationError(f"Missing URL path parameter: {param_name}")

            # Validate path params
            if validate is True:
                self._validate_url(sorted_required_path_params)

            for k, v in sorted_required_path_params.items():
                if path_params_safe_chars.get(k) is not None:
                    sorted_required_path_params[k] = quote(
                        str(v),
                        safe=path_params_safe_chars[k]
                    )
                else:
                    sorted_required_path_params[k] = quote(str(v))

            # Build full path
            # replace named placeholders with empty ones. e.g. {param} --> {}
            # Why? Because some endpoints have different names in their url path placeholders than in their parameter defenitions
            # e.g. path: {"v1/{+resourceName}/connections"}. e.g. param name: resourceName NOT +resourceName
            self._method_specs["path"] = re.sub(
                r"\{(.*?)\}", r"{}", self._method_specs["path"]
            )
            return base_url + self["path"].format(
                *list(sorted_required_path_params.values())
            )
        else:
            return base_url + self["path"]

    def _build_upload_media(self, upload_file, qualified_url, validate, fallback_url, pipe_from):
        if not self["supportsMediaUpload"]:
            if validate:
                raise ValidationError(
                    "upload_file was provided while method doesn't support media upload"
                )
            # This will probably not work, but will return a mediaupload object anyway
            return MediaUpload(upload_file, upload_path=fallback_url)

        # If resumable, create resumable object
        resumable = (
            self._build_resumeable_media(qualified_url)
            if _safe_getitem(
                self._method_specs, "mediaUpload", "protocols", "resumable"
            )
            else None
        )

        # Create MediaUpload object and pass it the resumable object we just created
        media_upload_url_base = self._root_url + "upload/" + self._service_path
        media_upload_url = qualified_url.replace(self._base_url, media_upload_url_base)
        max_size = self._convert_str_size_to_int(
            _safe_getitem(self._method_specs, "mediaUpload", "maxSize")
        )
        mime_range = _safe_getitem(self._method_specs, "mediaUpload", "accept")
        multipart = self["mediaUpload"]["protocols"]["simple"].get("multipart", True)

        return MediaUpload(
            upload_file,
            pipe_from=pipe_from,
            upload_path=media_upload_url,
            max_size=max_size,
            mime_range=mime_range,
            multipart=multipart,
            resumable=resumable,
            validate=validate,
        )

    def _build_resumeable_media(self, qualified_url):
        resumable_url_base = self._root_url + "resumable/upload/" + self._service_path
        resumable_url = qualified_url.replace(self._base_url, resumable_url_base)
        multipart = self["mediaUpload"]["protocols"]["resumable"].get("multipart", True)
        return ResumableUpload(multipart=multipart, upload_path=resumable_url)

    @staticmethod
    def _convert_str_size_to_int(size):
        """Convert a string media size, such as 10GB or 3TB into an integer.

        Args:
            size: (str): e.g. such as 2MB or 7GB.

        Returns:
            The size as an integer value.
        """
        if size is None:
            return None
        if len(size) < 2:
            return 0
        units = size[-2:].upper()
        bit_shift = MEDIA_SIZE_BIT_SHIFTS.get(units)
        if bit_shift is not None:
            return int(size[:-2]) << bit_shift
        else:
            return int(size)

    def _validate_url(self, sorted_required_path_params):
        for path_param_name, path_param_info in sorted_required_path_params.items():
            self._validate(
                instance=path_param_info,
                schema=self.parameters[path_param_name],
                schema_name=path_param_name,
            )

    def _validate_body(self, req):
        request_schema = self._method_specs.get("request")
        if request_schema is not None:
            schema_name = "Request Body"
            if "$ref" in request_schema:
                schema_name = request_schema["$ref"]
                request_schema = self._schemas[schema_name]
            self._validate(req, request_schema, schema_name=schema_name)
        else:
            raise ValidationError(
                "Request body should've been validated, but wasn't because the method doesn't accept a JSON body"
            )

    def __contains__(self, item):
        return item in self.parameters

    def __str__(self):
        return self["id"] + " method @ " + self._base_url

    def __repr__(self):
        return self.__str__()

    def __len__(self):
        return len(self.required_parameters) if self.required_parameters else 0


class Resource:
    def __init__(
        self,
        name,
        resource_specs,
        global_parameters,
        schemas,
        root_url,
        service_path,
        batch_path,
        validate,
    ):
        self.name = name
        self._resource_specs = resource_specs
        self._global_parameters = global_parameters
        self._schemas = schemas
        self._root_url = root_url
        self._service_path = service_path
        self._batch_path = batch_path
        self._validate = validate

    @property
    def methods_available(self) -> List[str]:
        """
        Returns the names of the methods that this resource provides
        """
        return [k for k, v in self["methods"].items()] if self["methods"] else []

    @property
    def resources_available(self) -> List[str]:
        """
        Returns the names of the nested resources in this resource
        """
        return [k for k, v in self["resources"].items()] if self["resources"] else []

    def _get_resource(self, resource_name):
        return Resource(
            name=resource_name,
            resource_specs=self["resources"][resource_name],
            global_parameters=self._global_parameters,
            schemas=self._schemas,
            root_url=self._root_url,
            service_path=self._service_path,
            batch_path=self._batch_path,
            validate=self._validate,
        )

    def _get_method(self, method_name):
        return Method(
            name=method_name,
            method_specs=self["methods"][method_name],
            global_parameters=self._global_parameters,
            schemas=self._schemas,
            root_url=self._root_url,
            service_path=self._service_path,
            batch_path=self._batch_path,
            validate=self._validate,
        )

    def __str__(self):
        return self.name + " resource @ " + self._root_url + self._service_path

    def __repr__(self):
        return self.__str__()

    def __call__(self):
        raise TypeError(
            "Only methods are callables, not resources."
            " e.g. api.resource.list() NOT api.resource().list()"
        )

    def __len__(self):
        return len(self.methods_available)

    def __contains__(self, item):
        return (item in self.methods_available) or (item in self.resources_available)

    def __getitem__(self, k):
        return self._resource_specs.get(k)

    def __getattr__(self, method_or_resource):
        """
        Returns either a method or a nested resource

        Arguments:

            method_or_resource: Name of the method or resource desired.

        Returns:

            aiogoogle.resource.Resource, aiogoogle.resource.Methods: A Resource or a Method


        Note:
            
            This method will first check in nested resources then will check in methods.

        Raises:

            AttributeError:
        """
        # 1. Search in nested resources
        if method_or_resource in self.resources_available:
            return self._get_resource(method_or_resource)
        # 2. Search in methods
        elif method_or_resource in self.methods_available:
            return self._get_method(method_or_resource)
        else:
            raise AttributeError(
                f"""Resource/Method {method_or_resource} doesn't exist.
                                 Check: https://developers.google.com/ for more info. 
                                 \nAvailable resources are: 
                                 {self.resources_available}\n
                                 Available methods are {self.methods_available}"""
            )


class GoogleAPI:
    """
    Creetes a representation of Google API given a discovery document
    
    Arguments:

        discovery_document (dict): A discovery document

        validate (bool): Set this to True to use this lib's built in parameter validation logic. Note that you shouldn't rely on this for critical user input validation.
    """

    def __init__(self, discovery_document, validate=False):
        self.discovery_document = self._add_extra_query_param_definitions(
            discovery_document
        )
        self._validate = validate

    def _add_extra_query_param_definitions(self, discovery_document):
        """ Adds extra parameters that aren't explicitly defined in discovery docuemnts
            i.e. "trace", "pp", "strict"
        """
        extra_params = {
            param: STACK_QUERY_PARAMETER_DEFAULT_VALUE
            for param in STACK_QUERY_PARAMETERS
        }
        if discovery_document.get("parameters"):
            discovery_document["parameters"] = {
                **discovery_document["parameters"],
                **extra_params,
            }
        else:
            discovery_document["parameters"] = extra_params
        return discovery_document

    @property
    def methods_available(self) -> List[str]:
        """
        Returns names of the methods provided by this resource
        """
        return [k for k, v in self["methods"].items()] if self["methods"] else []

    @property
    def resources_available(self) -> List[str]:
        """
        Returns names of the resources in a given API if any
        """
        return [k for k, v in self["resources"].items()] if self["resources"] else []

    def _get_resource(self, resource_name):
        return Resource(
            name=resource_name,
            resource_specs=self["resources"][resource_name],
            global_parameters=self["parameters"],
            schemas=self["schemas"]
            or {},  # josnschema validator will fail if schemas isn't a dict
            root_url=self["rootUrl"],
            service_path=self["servicePath"],
            batch_path=self["batchPath"],
            validate=self._validate,
        )

    def _get_method(self, method_name):
        return Method(
            name=method_name,
            method_specs=self["methods"][method_name],
            global_parameters=self["parameters"],
            schemas=self["schemas"]
            or {},  # josnschema validator will fail if schemas isn't a dict
            root_url=self["rootUrl"],
            service_path=self["servicePath"],
            batch_path=self["batchPath"],
            validate=self._validate,
        )

    def __getattr__(self, method_or_resource) -> Resource:
        """
        Returns resources from an API

        Note:
            
            This method will first check in resources then will check in methods.

        Arguments:

            method_or_resource (str): name of the top level method or resource

        Example:

            ::

                >>> google_service = GoogleAPI(google_service_discovery_doc)
                >>> google_service.user

                user resource @ google_service.com/api/v1/

        Returns:

            aiogoogle.resource.Resource, aiogoogle.resource.Method: A Resource or a Method

        Raises:

            AttributeError:
        """
        if method_or_resource in self.resources_available:
            return self._get_resource(method_or_resource)
        elif method_or_resource in self.methods_available:
            return self._get_method(method_or_resource)
        else:
            documentation_link = (
                self.discovery_document.get("documentationLink")
                or "https://developers.google.com/"
            )
            raise AttributeError(
                f"""Resource/Method {method_or_resource} doesn't exist.
                                 Check: {documentation_link} for more info. 
                                 \nAvailable resources are: 
                                 {self.resources_available}\n
                                 Available methods are {self.methods_available}"""
            )

    def __getitem__(self, k):
        """
        Returns items from the discovery document

        Example:

            ::

                >>> google_service = GoogleAPI(google_books_discovery_doc)
                >>> google_service['name']

                'books'

                >>> google_service['id']

                'books:v1'

                >>> google_service['version']

                'v1'

                >>> google_service['documentationLink']

                'https://developers.google.com/books/docs/v1/getting_started'                
                
                >>> google_service['oauth2']['scopes']

                https://www.googleapis.com/auth/books: {
                    
                    description: "Manage your books"
                
                }

        Returns:

            dict: Discovery Document Item
        """
        return self.discovery_document.get(k)

    def __contains__(self, name):
        return (name in self.resources_available) or (name in self.methods_available)

    def __repr__(self):
        labels = f'\nLabels:\n{self["labels"]}' if self["labels"] is not None else ""
        return (
            self.discovery_document["name"]
            + "-"
            + self.discovery_document["version"]
            + " API @ "
            + self["rootUrl"]
            + self["servicePath"]
            + labels
        )

    def __str__(self):
        return self.__repr__()

    def __len__(self):
        return len(self.resources_available) + len(self.methods_available)

    def __call__(self):
        raise TypeError(
            "Only methods are callables, not resources."
            " e.g. client.resources.user.list() NOT client.resources().user.list()"
        )



DISCOVERY_SERVICE_V1_DISCOVERY_DOC = {
    "basePath": "/discovery/v1/",
    "baseUrl": "https://www.googleapis.com/discovery/v1/",
    "batchPath": "batch/discovery/v1",
    "description": "Provides information about other Google APIs, such as what APIs are available, the resource, and method details for each API.",
    "discoveryVersion": "v1",
    "documentationLink": "https://developers.google.com/discovery/",
    "etag": '"J3WqvAcMk4eQjJXvfSI4Yr8VouA/kUro_lCKdlL_SgLaWY4jaDKtqsk"',
    "icons": {
        "x16": "http://www.google.com/images/icons/feature/filing_cabinet_search-g16.png",
        "x32": "http://www.google.com/images/icons/feature/filing_cabinet_search-g32.png",
    },
    "id": "discovery:v1",
    "kind": "discovery#restDescription",
    "name": "discovery",
    "ownerDomain": "google.com",
    "ownerName": "Google",
    "parameters": {
        "alt": {
            "default": "json",
            "description": "Data format for the response.",
            "enum": ["json"],
            "enumDescriptions": ["Responses with Content-Type of application/json"],
            "location": "query",
            "type": "string",
        },
        "fields": {
            "description": "Selector specifying which fields to include in a partial response.",
            "location": "query",
            "type": "string",
        },
        "key": {
            "description": "API key. Your API key identifies your project and provides you with API access, quota, and reports. Required unless you provide an OAuth 2.0 token.",
            "location": "query",
            "type": "string",
        },
        "oauth_token": {
            "description": "OAuth 2.0 token for the current user.",
            "location": "query",
            "type": "string",
        },
        "prettyPrint": {
            "default": "true",
            "description": "Returns response with indentations and line breaks.",
            "location": "query",
            "type": "boolean",
        },
        "quotaUser": {
            "description": "An opaque string that represents a user for quota purposes. Must not exceed 40 characters.",
            "location": "query",
            "type": "string",
        },
        "userIp": {
            "description": "Deprecated. Please use quotaUser instead.",
            "location": "query",
            "type": "string",
        },
    },
    "protocol": "rest",
    "resources": {
        "apis": {
            "methods": {
                "getRest": {
                    "description": "Retrieve the description of a particular version of an api.",
                    "httpMethod": "GET",
                    "id": "discovery.apis.getRest",
                    "parameterOrder": ["api", "version"],
                    "parameters": {
                        "api": {
                            "description": "The name of the API.",
                            "location": "path",
                            "required": True,
                            "type": "string",
                        },
                        "version": {
                            "description": "The version of the API.",
                            "location": "path",
                            "required": True,
                            "type": "string",
                        },
                    },
                    "path": "apis/{api}/{version}/rest",
                    "response": {"$ref": "RestDescription"},
                },
                "list": {
                    "description": "Retrieve the list of APIs supported at this endpoint.",
                    "httpMethod": "GET",
                    "id": "discovery.apis.list",
                    "parameters": {
                        "name": {
                            "description": "Only include APIs with the given name.",
                            "location": "query",
                            "type": "string",
                        },
                        "preferred": {
                            "default": "false",
                            "description": "Return only the preferred version of an API.",
                            "location": "query",
                            "type": "boolean",
                        },
                    },
                    "path": "apis",
                    "response": {"$ref": "DirectoryList"},
                },
            }
        }
    },
    "revision": "20181107",
    "rootUrl": "https://www.googleapis.com/",
    "schemas": {
        "DirectoryList": {
            "id": "DirectoryList",
            "properties": {
                "discoveryVersion": {
                    "default": "v1",
                    "description": "Indicate the version of the Discovery API used to generate this doc.",
                    "type": "string",
                },
                "items": {
                    "description": "The individual directory entries. One entry per api/version pair.",
                    "items": {
                        "properties": {
                            "description": {
                                "description": "The description of this API.",
                                "type": "string",
                            },
                            "discoveryLink": {
                                "description": "A link to the discovery document.",
                                "type": "string",
                            },
                            "discoveryRestUrl": {
                                "description": "The URL for the discovery REST document.",
                                "type": "string",
                            },
                            "documentationLink": {
                                "description": "A link to human readable documentation for the API.",
                                "type": "string",
                            },
                            "icons": {
                                "description": "Links to 16x16 and 32x32 icons representing the API.",
                                "properties": {
                                    "x16": {
                                        "description": "The URL of the 16x16 icon.",
                                        "type": "string",
                                    },
                                    "x32": {
                                        "description": "The URL of the 32x32 icon.",
                                        "type": "string",
                                    },
                                },
                                "type": "object",
                            },
                            "id": {
                                "description": "The id of this API.",
                                "type": "string",
                            },
                            "kind": {
                                "default": "discovery#directoryItem",
                                "description": "The kind for this response.",
                                "type": "string",
                            },
                            "labels": {
                                "description": "Labels for the status of this API, such as labs or deprecated.",
                                "items": {"type": "string"},
                                "type": "array",
                            },
                            "name": {
                                "description": "The name of the API.",
                                "type": "string",
                            },
                            "preferred": {
                                "description": "True if this version is the preferred version to use.",
                                "type": "boolean",
                            },
                            "title": {
                                "description": "The title of this API.",
                                "type": "string",
                            },
                            "version": {
                                "description": "The version of the API.",
                                "type": "string",
                            },
                        },
                        "type": "object",
                    },
                    "type": "array",
                },
                "kind": {
                    "default": "discovery#directoryList",
                    "description": "The kind for this response.",
                    "type": "string",
                },
            },
            "type": "object",
        },
        "JsonSchema": {
            "id": "JsonSchema",
            "properties": {
                "$ref": {
                    "description": 'A reference to another schema. The value of this property is the "id" of another schema.',
                    "type": "string",
                },
                "additionalProperties": {
                    "$ref": "JsonSchema",
                    "description": "If this is a schema for an object, this property is the schema for any additional properties with dynamic keys on "
                    "this object.",
                },
                "annotations": {
                    "description": "Additional information about this property.",
                    "properties": {
                        "required": {
                            "description": "A list of methods for which this property is required on requests.",
                            "items": {"type": "string"},
                            "type": "array",
                        }
                    },
                    "type": "object",
                },
                "default": {
                    "description": "The default value of this property (if one exists).",
                    "type": "string",
                },
                "description": {
                    "description": "A description of this object.",
                    "type": "string",
                },
                "enum": {
                    "description": "Values this parameter may take (if it is an enum).",
                    "items": {"type": "string"},
                    "type": "array",
                },
                "enumDescriptions": {
                    "description": 'The descriptions for the enums. Each position maps to the corresponding value in the "enum" array.',
                    "items": {"type": "string"},
                    "type": "array",
                },
                "format": {
                    "description": "An additional regular expression or key that helps constrain the value. For more details see: "
                    "http://tools.ietf.org/html/draft-zyp-json-schema-03#section-5.23",
                    "type": "string",
                },
                "id": {
                    "description": "Unique identifier for this schema.",
                    "type": "string",
                },
                "items": {
                    "$ref": "JsonSchema",
                    "description": "If this is a schema for an array, this property is the schema for each element in the array.",
                },
                "location": {
                    "description": "Whether this parameter goes in the query or the path for REST requests.",
                    "type": "string",
                },
                "maximum": {
                    "description": "The maximum value of this parameter.",
                    "type": "string",
                },
                "minimum": {
                    "description": "The minimum value of this parameter.",
                    "type": "string",
                },
                "pattern": {
                    "description": "The regular expression this parameter must conform to. Uses Java 6 regex format: "
                    "http://docs.oracle.com/javase/6/docs/api/java/util/regex/Pattern.html",
                    "type": "string",
                },
                "properties": {
                    "additionalProperties": {
                        "$ref": "JsonSchema",
                        "description": "A single property of this object. The value is itself a JSON Schema object describing this "
                        "property.",
                    },
                    "description": "If this is a schema for an object, list the schema for each property of this object.",
                    "type": "object",
                },
                "readOnly": {
                    "description": "The value is read-only, generated by the service. The value cannot be modified by the client. If the value is included in a "
                    "POST, PUT, or PATCH request, it is ignored by the service.",
                    "type": "boolean",
                },
                "repeated": {
                    "description": "Whether this parameter may appear multiple times.",
                    "type": "boolean",
                },
                "required": {
                    "description": "Whether the parameter is required.",
                    "type": "boolean",
                },
                "type": {
                    "description": "The value type for this schema. A list of values can be found here: "
                    "http://tools.ietf.org/html/draft-zyp-json-schema-03#section-5.1",
                    "type": "string",
                },
                "variant": {
                    "description": "In a variant data type, the value of one property is used to determine how to interpret the entire entity. Its value must exist "
                    "in a map of descriminant values to schema names.",
                    "properties": {
                        "discriminant": {
                            "description": "The name of the type discriminant property.",
                            "type": "string",
                        },
                        "map": {
                            "description": "The map of discriminant value to schema to use for parsing..",
                            "items": {
                                "properties": {
                                    "$ref": {"type": "string"},
                                    "type_value": {"type": "string"},
                                },
                                "type": "object",
                            },
                            "type": "array",
                        },
                    },
                    "type": "object",
                },
            },
            "type": "object",
        },
        "RestDescription": {
            "id": "RestDescription",
            "properties": {
                "auth": {
                    "description": "Authentication information.",
                    "properties": {
                        "oauth2": {
                            "description": "OAuth 2.0 authentication information.",
                            "properties": {
                                "scopes": {
                                    "additionalProperties": {
                                        "description": "The scope value.",
                                        "properties": {
                                            "description": {
                                                "description": "Description of "
                                                "scope.",
                                                "type": "string",
                                            }
                                        },
                                        "type": "object",
                                    },
                                    "description": "Available OAuth 2.0 scopes.",
                                    "type": "object",
                                }
                            },
                            "type": "object",
                        }
                    },
                    "type": "object",
                },
                "basePath": {
                    "description": "[DEPRECATED] The base path for REST requests.",
                    "type": "string",
                },
                "baseUrl": {
                    "description": "[DEPRECATED] The base URL for REST requests.",
                    "type": "string",
                },
                "batchPath": {
                    "description": "The path for REST batch requests.",
                    "type": "string",
                },
                "canonicalName": {
                    "description": "Indicates how the API name should be capitalized and split into various parts. Useful for generating pretty class "
                    "names.",
                    "type": "string",
                },
                "description": {
                    "description": "The description of this API.",
                    "type": "string",
                },
                "discoveryVersion": {
                    "default": "v1",
                    "description": "Indicate the version of the Discovery API used to generate this doc.",
                    "type": "string",
                },
                "documentationLink": {
                    "description": "A link to human readable documentation for the API.",
                    "type": "string",
                },
                "etag": {
                    "description": "The ETag for this response.",
                    "readOnly": True,
                    "type": "string",
                },
                "exponentialBackoffDefault": {
                    "description": "Enable exponential backoff for suitable methods in the generated clients.",
                    "type": "boolean",
                },
                "features": {
                    "description": "A list of supported features for this API.",
                    "items": {"type": "string"},
                    "type": "array",
                },
                "icons": {
                    "description": "Links to 16x16 and 32x32 icons representing the API.",
                    "properties": {
                        "x16": {
                            "description": "The URL of the 16x16 icon.",
                            "type": "string",
                        },
                        "x32": {
                            "description": "The URL of the 32x32 icon.",
                            "type": "string",
                        },
                    },
                    "type": "object",
                },
                "id": {"description": "The ID of this API.", "type": "string"},
                "kind": {
                    "default": "discovery#restDescription",
                    "description": "The kind for this response.",
                    "type": "string",
                },
                "labels": {
                    "description": "Labels for the status of this API, such as labs or deprecated.",
                    "items": {"type": "string"},
                    "type": "array",
                },
                "methods": {
                    "additionalProperties": {
                        "$ref": "RestMethod",
                        "description": "An individual method description.",
                    },
                    "description": "API-level methods for this API.",
                    "type": "object",
                },
                "name": {"description": "The name of this API.", "type": "string"},
                "ownerDomain": {
                    "description": "The domain of the owner of this API. Together with the ownerName and a packagePath values, this can be used to "
                    "generate a library for this API which would have a unique fully qualified name.",
                    "type": "string",
                },
                "ownerName": {
                    "description": "The name of the owner of this API. See ownerDomain.",
                    "type": "string",
                },
                "packagePath": {
                    "description": "The package of the owner of this API. See ownerDomain.",
                    "type": "string",
                },
                "parameters": {
                    "additionalProperties": {
                        "$ref": "JsonSchema",
                        "description": "Description of a single parameter.",
                    },
                    "description": "Common parameters that apply across all apis.",
                    "type": "object",
                },
                "protocol": {
                    "default": "rest",
                    "description": "The protocol described by this document.",
                    "type": "string",
                },
                "resources": {
                    "additionalProperties": {
                        "$ref": "RestResource",
                        "description": "An individual resource description. Contains methods and sub-resources related to this "
                        "resource.",
                    },
                    "description": "The resources in this API.",
                    "type": "object",
                },
                "revision": {
                    "description": "The version of this API.",
                    "type": "string",
                },
                "rootUrl": {
                    "description": "The root URL under which all API services live.",
                    "type": "string",
                },
                "schemas": {
                    "additionalProperties": {
                        "$ref": "JsonSchema",
                        "description": "An individual schema description.",
                    },
                    "description": "The schemas for this API.",
                    "type": "object",
                },
                "servicePath": {
                    "description": "The base path for all REST requests.",
                    "type": "string",
                },
                "title": {"description": "The title of this API.", "type": "string"},
                "version": {
                    "description": "The version of this API.",
                    "type": "string",
                },
                "version_module": {"type": "boolean"},
            },
            "type": "object",
        },
        "RestMethod": {
            "id": "RestMethod",
            "properties": {
                "description": {
                    "description": "Description of this method.",
                    "type": "string",
                },
                "etagRequired": {
                    "description": "Whether this method requires an ETag to be specified. The ETag is sent as an HTTP If-Match or If-None-Match header.",
                    "type": "boolean",
                },
                "httpMethod": {
                    "description": "HTTP method used by this method.",
                    "type": "string",
                },
                "id": {
                    "description": "A unique ID for this method. This property can be used to match methods between different versions of Discovery.",
                    "type": "string",
                },
                "mediaUpload": {
                    "description": "Media upload parameters.",
                    "properties": {
                        "accept": {
                            "description": "MIME Media Ranges for acceptable media uploads to this method.",
                            "items": {"type": "string"},
                            "type": "array",
                        },
                        "maxSize": {
                            "description": 'Maximum size of a media upload, such as "1MB", "2GB" or "3TB".',
                            "type": "string",
                        },
                        "protocols": {
                            "description": "Supported upload protocols.",
                            "properties": {
                                "resumable": {
                                    "description": "Supports the Resumable Media Upload protocol.",
                                    "properties": {
                                        "multipart": {
                                            "default": "true",
                                            "description": "True if this endpoint supports "
                                            "uploading multipart media.",
                                            "type": "boolean",
                                        },
                                        "path": {
                                            "description": "The URI path to be used for upload. "
                                            "Should be used in conjunction with the "
                                            "basePath property at the api-level.",
                                            "type": "string",
                                        },
                                    },
                                    "type": "object",
                                },
                                "simple": {
                                    "description": "Supports uploading as a single HTTP request.",
                                    "properties": {
                                        "multipart": {
                                            "default": "true",
                                            "description": "True if this endpoint supports upload "
                                            "multipart media.",
                                            "type": "boolean",
                                        },
                                        "path": {
                                            "description": "The URI path to be used for upload. Should "
                                            "be used in conjunction with the basePath "
                                            "property at the api-level.",
                                            "type": "string",
                                        },
                                    },
                                    "type": "object",
                                },
                            },
                            "type": "object",
                        },
                    },
                    "type": "object",
                },
                "parameterOrder": {
                    "description": "Ordered list of required parameters, serves as a hint to clients on how to structure their method signatures. The array "
                    'is ordered such that the "most-significant" parameter appears first.',
                    "items": {"type": "string"},
                    "type": "array",
                },
                "parameters": {
                    "additionalProperties": {
                        "$ref": "JsonSchema",
                        "description": "Details for a single parameter in this method.",
                    },
                    "description": "Details for all parameters in this method.",
                    "type": "object",
                },
                "path": {
                    "description": "The URI path of this REST method. Should be used in conjunction with the basePath property at the api-level.",
                    "type": "string",
                },
                "request": {
                    "description": "The schema for the request.",
                    "properties": {
                        "$ref": {
                            "description": "Schema ID for the request schema.",
                            "type": "string",
                        },
                        "parameterName": {
                            "description": "parameter name.",
                            "type": "string",
                        },
                    },
                    "type": "object",
                },
                "response": {
                    "description": "The schema for the response.",
                    "properties": {
                        "$ref": {
                            "description": "Schema ID for the response schema.",
                            "type": "string",
                        }
                    },
                    "type": "object",
                },
                "scopes": {
                    "description": "OAuth 2.0 scopes applicable to this method.",
                    "items": {"type": "string"},
                    "type": "array",
                },
                "supportsMediaDownload": {
                    "description": "Whether this method supports media downloads.",
                    "type": "boolean",
                },
                "supportsMediaUpload": {
                    "description": "Whether this method supports media uploads.",
                    "type": "boolean",
                },
                "supportsSubscription": {
                    "description": "Whether this method supports subscriptions.",
                    "type": "boolean",
                },
                "useMediaDownloadService": {
                    "description": 'Indicates that downloads from this method should use the download service URL (i.e. "/download"). Only applies '
                    "if the method supports media download.",
                    "type": "boolean",
                },
            },
            "type": "object",
        },
        "RestResource": {
            "id": "RestResource",
            "properties": {
                "methods": {
                    "additionalProperties": {
                        "$ref": "RestMethod",
                        "description": "Description for any methods on this resource.",
                    },
                    "description": "Methods on this resource.",
                    "type": "object",
                },
                "resources": {
                    "additionalProperties": {
                        "$ref": "RestResource",
                        "description": "Description for any sub-resources on this resource.",
                    },
                    "description": "Sub-resources on this resource.",
                    "type": "object",
                },
            },
            "type": "object",
        },
    },
    "servicePath": "discovery/v1/",
    "title": "APIs Discovery Service",
    "version": "v1",
}



from setuptools import setup, find_packages
from distutils.util import convert_path

import os
from os import listdir
from os.path import isfile, join


main_ns = {}
ver_path = convert_path('aiogoogle/__version__.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

mypath = os.path.dirname(os.path.abspath(__file__))
print([f for f in listdir(mypath) if isfile(join(mypath, f))])


with open("requirements.txt", "r") as f:
    requirements = f.read().splitlines()

with open("test_requirements.txt", "r") as f:
    test_requirements = f.read().splitlines()

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name=main_ns['__name__'],
    version=main_ns['__version__'],
    author=main_ns['__author__'],
    author_email=main_ns['__author_email__'],
    license=main_ns['__license__'],
    description=main_ns['__description__'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[requirements, 'contextvars;python_version<"3.7"'],
    tests_require=test_requirements,
    url=main_ns["__url__"],
    packages=find_packages(exclude=['tests*']),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ],
    extras_require={"curio_asks": ["asks", "curio"], "trio_asks": ["asks", "trio"]},
)
